package pogr;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.util.Random;

/**
 * Class with methods for dithering
 * @author KarelSimanek
 */

public class Dither {
    
    /**
     * 8 color palette
     */
    static C3[] palette = new C3[] {
            new C3(  0,   0,   0),
            new C3(  0,   0, 255),
            new C3(  0, 255,   0),
            new C3(  0, 255, 255),
            new C3(255,   0,   0),
            new C3(255,   0, 255),
            new C3(255, 255,   0),
            new C3(255, 255, 255)
            }; 
    /**
     * 2 colors palette
     */
            static C3[] palette2 = new C3[] {
            new C3(  0,   0,   0),
            new C3(255, 255, 255)
            };  
    /**
     * FloydSteinberg dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Floyd-Steinberg algorithm
     */
    public static BufferedImage FloydSteinberg(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
        }

          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 C3 oldColor = d[y][x];
                 C3 newColor = findClosestPaletteColor(oldColor, palette);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());

                 C3 err = oldColor.sub(newColor);

                if (x+1 < w) {
                     d[y  ][x+1] = d[y  ][x+1].add(err.mul(7./16));
                 }
                if (x-1>=0 && y+1<h) {
                     d[y+1][x-1] = d[y+1][x-1].add(err.mul(3./16));
                 }
                if (y+1 < h) {
                     d[y+1][x  ] = d[y+1][x  ].add(err.mul(5./16));
                 }
                if (x+1<w && y+1<h) {
                     d[y+1][x+1] = d[y+1][x+1].add(err.mul(1./16));
                 }
            }
         }
    return newImg;
    }
    
    /**
     * Atkinson dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Atkinson algorithm
     */
    public static BufferedImage Atkinson(BufferedImage img)
    {
           int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
        }

          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 C3 oldColor = d[y][x];
                 C3 newColor = findClosestPaletteColor(oldColor, palette);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());

                 C3 err = oldColor.sub(newColor);

                if (x+1 < w) {
                     d[y  ][x+1] = d[y  ][x+1].add(err.mul(1./8));
                 }
                if (x+2 < w) {
                     d[y  ][x+2] = d[y  ][x+2].add(err.mul(1./8));
                 }
                if (x-1>=0 && y+1<h) {
                     d[y+1][x-1] = d[y+1][x-1].add(err.mul(1./8));
                 }
                if (y+1 < h) {
                     d[y+1][x  ] = d[y+1][x  ].add(err.mul(1./8));
                 }
                if (x+1<w && y+1<h) {
                     d[y+1][x+1] = d[y+1][x+1].add(err.mul(1./8));
                 }
                if (y+2 < h) {
                     d[y+2  ][x] = d[y+2 ][x].add(err.mul(1./8));
                 }
            }
         }
    return newImg;
    }
    
    /**
     * Stucki dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Stucki algorithm
     */
    public static BufferedImage Stucki(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
        }

          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 C3 oldColor = d[y][x];
                 C3 newColor = findClosestPaletteColor(oldColor, palette);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());

                 C3 err = oldColor.sub(newColor);

                if (x+1 < w) {
                     d[y  ][x+1] = d[y  ][x+1].add(err.mul(7./42));
                 }
                if (x+2 < w) {
                     d[y  ][x+2] = d[y  ][x+2].add(err.mul(5./42));
                 }
                if (x-2>=0 && y+1<h) {
                     d[y+1][x-2] = d[y+1][x-2].add(err.mul(2./42));
                 }                
                if (x-1>=0 && y+1<h) {
                     d[y+1][x-1] = d[y+1][x-1].add(err.mul(4./42));
                 }
                if (y+1 < h) {
                     d[y+1  ][x] = d[y+1][x].add(err.mul(8./42));
                 }
                if (x+1<w && y+1<h) {
                     d[y+1][x+1] = d[y+1][x+1].add(err.mul(4./42));
                 }
                if (x+2<w && y+1<h) {
                     d[y+1][x+2] = d[y+1][x+2].add(err.mul(2./42));
                 }
                if (x-2>=0 && y+2<h) {
                     d[y+2][x-2] = d[y+2][x-2].add(err.mul(1./42));
                 }
                if (x-1>=0 && y+2<h) {
                     d[y+2][x-1] = d[y+2][x-1].add(err.mul(2./42));
                 }
                if (y+2 < h) {
                     d[y+2  ][x] = d[y+2 ][x].add(err.mul(4./42));
                 }
                if (x+1<w && y+2<h) {
                     d[y+2][x+1] = d[y+2][x+1].add(err.mul(2./42));
                 }
                if (x+2<w && y+2<h) {
                     d[y+2][x+2] = d[y+2][x+2].add(err.mul(1./42));
                 }
            }
         }
    return newImg;
    }
    
    /**
     * 
     * @param img Input BufferedImage
     * @param tresholdRed red treshold value
     * @param tresholdGreen green treshold value
     * @param tresholdBlue blue treshold value
     * @return Output BufferedImage with applied Tresholding algorithm
     */
    public static BufferedImage Tresholding(BufferedImage img, int tresholdRed, int tresholdGreen, int tresholdBlue)
    {
            
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());
            
            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
        } 
            
          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 int red = (d[y][x].r > tresholdRed) ? 255 : 0;
                 int green = (d[y][x].g > tresholdGreen) ? 255 : 0;
                 int blue = (d[y][x].b > tresholdBlue) ? 255 : 0;
                 C3 newColor = new C3(red,green,blue);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());
            }
         }
    return newImg;
    }   
    
    /**
     * Stucki dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Clustered 4x4 algorithm
     */
    public static BufferedImage Clustered_4x4(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
            }
            
          double[] bayer = {0.7500,0.3750,0,6250,0.2500,0.0625,1.000,0.8750,04375,0.5000,0.8125,0.9375,0.1250,0.1875,0.5625,0.3125,0.6875};          
          for (int i = 0; i<bayer.length; i++) { 
              bayer[i] = bayer[i]*255;
          }
            
          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 int red = (d[y][x].r >= bayer[(y%4*4+x%4)]) ? 255 : 0;
                 int green = (d[y][x].g >= bayer[(y%4*4+x%4)]) ? 255 : 0;
                 int blue = (d[y][x].b >= bayer[(y%4*4+x%4)]) ? 255 : 0;
                 C3 newColor = new C3(red,green,blue);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());
            }
         }
    return newImg;
    }
    /**
     * Stucki dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied RandomDithering algorithm
     */
    public static BufferedImage Random(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());
            
            Random r = new Random();
            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
        } 
            
          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 int random = r.nextInt(255);
                 int red = (d[y][x].r > random) ? 255 : 0;
                 int green = (d[y][x].g > random) ? 255 : 0;
                 int blue = (d[y][x].b > random) ? 255 : 0;
                 C3 newColor = new C3(red,green,blue);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());
            }
         }
    return newImg;
    }  
    
    /**
     * Stucki dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Bayer 8x8 algorithm
     */
    public static BufferedImage Bayer_8x8(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
            }
            
          int[] bayer = {1, 33, 9, 41, 3, 35, 11, 43, 49, 17, 57, 25, 51,19,59,27,13,45,5,37,15,47,7,39,64,29,53,21,63,31,55,23,4,36,12, 44, 2, 34, 10, 42, 52, 20, 60, 28, 50, 18, 58, 26, 16, 48, 8, 40, 14, 46, 6, 38, 64, 32, 56, 24, 62, 30, 54, 22};
          for (int i = 0; i<bayer.length; i++) {
            bayer[i] = bayer[i]*4-1;
          }
            
          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 int red = (d[y][x].r >= bayer[(y%8*8+x%8)]) ? 255 : 0;
                 int green = (d[y][x].g >= bayer[(y%8*8+x%8)]) ? 255 : 0;
                 int blue = (d[y][x].b >= bayer[(y%8*8+x%8)]) ? 255 : 0;
                 C3 newColor = new C3(red,green,blue);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());
            }
         }
    return newImg;
    } 
    
    /**
     * Stucki dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Bayer 4x4 algorithm
     */
    public static BufferedImage Bayer_4x4(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
            }
            
          double[] bayer = {0.1250,1.0000,0.1875,0.8125,0.6250,0.3750,0.6875,0.4375,0.2500,0.8750,0.0625,0.9375,0.7500,0.5000,0,5625,0.3125};
          for (int i = 0; i<bayer.length; i++) {
            bayer[i] = bayer[i]*255;
          }
            
          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 int red = (d[y][x].r >= bayer[(y%4*4+x%4)]) ? 255 : 0;
                 int green = (d[y][x].g >= bayer[(y%4*4+x%4)]) ? 255 : 0;
                 int blue = (d[y][x].b >= bayer[(y%4*4+x%4)]) ? 255 : 0;
                 C3 newColor = new C3(red,green,blue);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());
            }
         }
    return newImg;
    }   
    
    /**
     * Stucki dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Bayer 2x2 algorithm
     */
    public static BufferedImage Bayer_2x2(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
            }
            
          int[] bayer = {2,3,4,1};
          for (int i = 0; i<bayer.length; i++) {
            bayer[i] = bayer[i]*64-1;
          }
            
          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 int red = (d[y][x].r >= bayer[(y%2*2+x%2)]) ? 255 : 0;
                 int green = (d[y][x].g >= bayer[(y%2*2+x%2)]) ? 255 : 0;
                 int blue = (d[y][x].b >= bayer[(y%2*2+x%2)]) ? 255 : 0;
                 C3 newColor = new C3(red,green,blue);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());
            }
         }
    return newImg;
    } 
    
    /**
     * Stucki dithering algorithm
     * @param img Input BufferedImage
     * @return Output BufferedImage with applied Jarvis-Judice-Ninke algorithm
     */
    public static BufferedImage JarvisJudiceNinke(BufferedImage img)
    {
            int w = img.getWidth();
            int h = img.getHeight();
            
            BufferedImage newImg = new BufferedImage(w,h, img.getType());

            C3[][] d = new C3[h][w];

            for (int y = 0; y < h; y++) { 
                for (int x = 0; x < w; x++) {
                  d[y][x] = new C3(img.getRGB(x, y));
              }
        }

          for (int y = 0; y < h; y++) {
             for (int x = 0; x < w; x++) {
                 C3 oldColor = d[y][x];
                 C3 newColor = findClosestPaletteColor(oldColor, palette);
                 newImg.setRGB(x, y, newColor.toColor().getRGB());

                 C3 err = oldColor.sub(newColor);

                if (x+1 < w) {
                     d[y  ][x+1] = d[y  ][x+1].add(err.mul(7./48));
                 }
                if (x+2 < w) {
                     d[y  ][x+2] = d[y  ][x+2].add(err.mul(5./48));
                 }
                if (x-2 >= 0 && y+1 < h) {
                     d[y+1][x-2] = d[y + 1 ][x-2].add(err.mul(3./48));
                 }
                if (x-1>=0 && y+1<h) {
                     d[y+1][x-1] = d[y+1][x-1].add(err.mul(5./48));
                 }
                if (y+1 < h) {
                     d[y+1][x  ] = d[y+1][x  ].add(err.mul(7./48));
                 }
                if (x+1<w && y+1<h) {
                     d[y+1][x+1] = d[y+1][x+1].add(err.mul(5./48));
                 }
                if (x+2<w && y+1<h) {
                     d[y+1][x+2] = d[y+1][x+2].add(err.mul(3./48));
                 }
                if (x-2>=0 && y+2<h) {
                     d[y+2][x-2] = d[y+2][x-2].add(err.mul(1./48));
                 }
                if (x-1>=0 && y+2<h) {
                     d[y+2][x-1] = d[y+2][x-1].add(err.mul(3./48));
                 }
                if(y+2 < h){
                     d[y+2][x  ] = d[y+2][x  ].add(err.mul(5./48));
                }
                if(x+1<w && y+2 < h){
                     d[y+2][x+1  ] = d[y+2][x+1  ].add(err.mul(3./48));
                }
                if(x+2<w && y+2 < h){
                     d[y+2][x+2  ] = d[y+2][x+2  ].add(err.mul(1./48));
                }
            }
         }
    return newImg;
    }   
    
    /**
     * Auxiliary method for finding the closest pallette color (maybe redundant, for our palletes the rounding is sufficient)
     * implemented using Euclidean distances
     * @param c input color
     * @param palette output color
     * @return Closest Palette Color
     */

    private static C3 findClosestPaletteColor(C3 c, C3[] palette) {
    C3 closest = palette[0];
    
    for (C3 n : palette) {
            if (n.diff(c) < closest.diff(c)) {
                closest = n;
            }
    }
    return closest;
  }
    
    //Auxiliary class for colors, in order the simplify the work with colors
    static class C3 {
    int r, g, b;
    
    /**
     * Constructor method for setting the grayscale value from integer
     * @param c grayscale value
     */
    public C3(int c) {
      Color color = new Color(c);
      this.r = color.getRed();
      this.g = color.getGreen();
      this.b = color.getBlue();
    }
    /**
     * Constructor method for setting the rgb value from 3 integers
     * @param r red value
     * @param g green value
     * @param b blue value
     */
    public C3(int r, int g, int b) {
      this.r = r;
      this.g = g;
      this.b = b;
    }

    /**
     * Adding for colors
     * @param o color operand
     * @return new color gained by the adding
     */
    public C3 add(C3 o) {
      return new C3(r + o.r, g + o.g, b + o.b);
    }
    /**
     * Subtraction for colors
     * @param o color operand
     * @return new color gained by the subtraction
     */
    public C3 sub(C3 o) {
      return new C3(r - o.r, g - o.g, b - o.b);
    }
    
    /**
     * Multiplication of color by double
     * @param d double operand
     * @return new color gained by the multiplication
     */
    public C3 mul(double d) {
      return new C3((int) (d * r), (int) (d * g), (int) (d * b));
    }
    
    /**
     * Conversion this class to RGB
     * @return RGB value
     */
    public int toRGB() {
      return toColor().getRGB();
    }
    /**
     * Conversion this class to Color
     * @return Color class
     */
    public Color toColor() {
      return new Color(clamp(r), clamp(g), clamp(b));
    }
    
    /**
     * This method ensure the range between 0 and 255
     * @param c input integer value
     * @return corrected value
     */
    public int clamp(int c) {
      return Math.max(0, Math.min(255, c));
    }
    
    /**
     * Euclidan distance between colors
     * @param c input color
     * @return double value of the difference between two colors
     */
    public double diff(C3 c)
    {
        return Math.sqrt((c.r-r)*(c.r-r)+(c.g-g)*(c.g-g)+(c.b-b)*(c.b-b));
    }
  }
}



